"""This is a somewhat delicate package. It contains all registered components
and preconfigured templates.

Hence, it imports all of the components. To avoid cycles, no component should
import this in module scope."""

import logging
import traceback
import typing
from typing import Any, Dict, Optional, Text, Type

from convo.nlu.classifiers.diet_classifier import DIETClassifier
from convo.nlu.classifiers.fallback_classifier import FallbackClassifier
from convo.nlu.classifiers.keyword_intent_classifier import KeywordIntentClassifier
from convo.nlu.classifiers.mitie_intent_classifier import MitieIntentClassifier
from convo.nlu.classifiers.sklearn_intent_classifier import SklearnIntentClassifier
from convo.nlu.extractors.crf_entity_extractor import CRFEntityExtractor
from convo.nlu.extractors.duckling_entity_extractor import DucklingEntityExtractor
from convo.nlu.extractors.entity_synonyms import EntitySynonymMapper
from convo.nlu.extractors.mitie_entity_extractor import MitieEntityExtractor
from convo.nlu.extractors.spacy_entity_extractor import SpacyEntityExtractor
from convo.nlu.extractors.regex_entity_extractor import RegexEntityExtractor
from convo.nlu.featurizers.sparse_featurizer.lexical_syntactic_featurizer import (
    LexicalSyntacticFeaturizer,
)
from convo.nlu.featurizers.dense_featurizer.convert_featurizer import ConveRTFeaturizer
from convo.nlu.featurizers.dense_featurizer.mitie_featurizer import MitieFeaturizer
from convo.nlu.featurizers.dense_featurizer.spacy_featurizer import SpacyFeaturizer
from convo.nlu.featurizers.sparse_featurizer.count_vectors_featurizer import (
    CountVectorsFeaturizer,
)
from convo.nlu.featurizers.dense_featurizer.lm_featurizer import LanguageModelFeaturizer
from convo.nlu.featurizers.sparse_featurizer.regex_featurizer import RegexFeaturizer
from convo.nlu.model import Metadata
from convo.nlu.selectors.response_selector import ResponseSelector
from convo.nlu.tokenizers.convert_tokenizer import ConveRTTokenizer
from convo.nlu.tokenizers.jieba_tokenizer import JiebaTokenizer
from convo.nlu.tokenizers.mitie_tokenizer import MitieTokenizer
from convo.nlu.tokenizers.spacy_tokenizer import SpacyTokenizer
from convo.nlu.tokenizers.whitespace_tokenizer import WhitespaceTokenizer
from convo.nlu.tokenizers.lm_tokenizer import LanguageModelTokenizer
from convo.nlu.utils.mitie_utils import MitieNLP
from convo.nlu.utils.spacy_utils import SpacyNLP
from convo.nlu.utils.hugging_face.hf_transformers import HFTransformersNLP
from convo.shared.exceptions import ConvoException
import convo.shared.utils.common
import convo.shared.utils.io
import convo.utils.io
from convo.shared.constants import DOCS_URL_COMPONENTS

if typing.TYPE_CHECKING:
    from convo.nlu.components import Component
    from convo.nlu.config import ConvoNLUModelConfig

logger = logging.getLogger(__name__)


# Classes of all known components. If a new component should be added,
# its class name should be listed here.
component_classes = [
    # utils
    SpacyNLP,
    MitieNLP,
    HFTransformersNLP,
    # tokenizers
    MitieTokenizer,
    SpacyTokenizer,
    WhitespaceTokenizer,
    ConveRTTokenizer,
    JiebaTokenizer,
    LanguageModelTokenizer,
    # extractors
    SpacyEntityExtractor,
    MitieEntityExtractor,
    CRFEntityExtractor,
    DucklingEntityExtractor,
    EntitySynonymMapper,
    RegexEntityExtractor,
    # featurizers
    SpacyFeaturizer,
    MitieFeaturizer,
    RegexFeaturizer,
    LexicalSyntacticFeaturizer,
    CountVectorsFeaturizer,
    ConveRTFeaturizer,
    LanguageModelFeaturizer,
    # classifiers
    SklearnIntentClassifier,
    MitieIntentClassifier,
    KeywordIntentClassifier,
    DIETClassifier,
    FallbackClassifier,
    # selectors
    ResponseSelector,
]

# Mapping from a components name to its class to allow name based lookup.
registered_components = {c.name: c for c in component_classes}


class ComponentNotFoundException(ModuleNotFoundError, ConvoException):
    """Raised if a module referenced by name can not be imported."""

    pass


def get_component_class(component_name: Text) -> Type["Component"]:
    """Resolve component name to a registered components class."""

    if component_name == "DucklingHTTPExtractor":
        convo.shared.utils.io.raise_deprecation_warning(
            "The component 'DucklingHTTPExtractor' has been renamed to "
            "'DucklingEntityExtractor'. Update your pipeline to use "
            "'DucklingEntityExtractor'.",
            docs=DOCS_URL_COMPONENTS,
        )
        component_name = "DucklingEntityExtractor"

    if component_name not in registered_components:
        try:
            return convo.shared.utils.common.class_from_module_path(component_name)

        except (ImportError, AttributeError) as e:
            # when component_name is a path to a class but that path is invalid or
            # when component_name is a class name and not part of old_style_names

            is_path = "." in component_name

            if is_path:
                module_name, _, class_name = component_name.rpartition(".")
                if isinstance(e, ImportError):
                    exception_message = f"Failed to find module '{module_name}'."
                else:
                    # when component_name is a path to a class but the path does
                    # not contain that class
                    exception_message = (
                        f"The class '{class_name}' could not be "
                        f"found in module '{module_name}'."
                    )
            else:
                exception_message = (
                    f"Cannot find class '{component_name}' in global namespace. "
                    f"Please check that there is no typo in the class "
                    f"name and that you have imported the class into the global "
                    f"namespace."
                )

            raise ComponentNotFoundException(
                f"Failed to load the component "
                f"'{component_name}'. "
                f"{exception_message} Either your "
                f"pipeline configuration contains an error "
                f"or the module you are trying to import "
                f"is broken (e.g. the module is trying "
                f"to import a package that is not "
                f"installed). {traceback.format_exc()}"
            )

    return registered_components[component_name]


def load_component_by_meta(
    component_meta: Dict[Text, Any],
    model_dir: Text,
    metadata: Metadata,
    cached_component: Optional["Component"],
    **kwargs: Any,
) -> Optional["Component"]:
    """Resolves a component and calls its load method.

    Inits it based on a previously persisted model.
    """

    # try to get class name first, else create by name
    component_name = component_meta.get("class", component_meta["name"])
    component_class = get_component_class(component_name)
    return component_class.load(
        component_meta, model_dir, metadata, cached_component, **kwargs
    )


def create_component_by_config(
    component_config: Dict[Text, Any], config: "ConvoNLUModelConfig"
) -> Optional["Component"]:
    """Resolves a component and calls it's create method.

    Inits it based on a previously persisted model.
    """

    # try to get class name first, else create by name
    component_name = component_config.get("class", component_config["name"])
    component_class = get_component_class(component_name)
    return component_class.create(component_config, config)
